"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
exports.__esModule = true;
var dockerode_1 = __importDefault(require("dockerode"));
var bull_1 = __importDefault(require("bull"));
var Runner_1 = __importDefault(require("./Runner"));
var Builder_1 = __importDefault(require("./Builder"));
var logger_1 = __importDefault(require("./utils/logger"));
var Worker = /** @class */ (function () {
    function Worker(name, redis, options) {
        this.docker = new dockerode_1["default"]();
        this.runner = new Runner_1["default"](this.docker);
        this.builder = new Builder_1["default"](this.docker);
        this.queue = new bull_1["default"](name, redis);
        var opts = options || {};
        var folderPath = opts.folderPath, memory = opts.memory, CPUs = opts.CPUs;
        this.folderPath = folderPath || '/tmp/code-exec';
        this.memory = memory || 0;
        this.CPUs = CPUs || 0.5;
    }
    Worker.prototype.work = function (codeOptions) {
        return __awaiter(this, void 0, void 0, function () {
            var tag, result;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        tag = codeOptions.language.toLowerCase() + "-runner";
                        return [4 /*yield*/, this.runner.run({
                                tag: tag,
                                id: codeOptions.id,
                                code: codeOptions.code,
                                testCases: codeOptions.testCases,
                                folderPath: this.folderPath,
                                base64: codeOptions.base64 || false,
                                language: codeOptions.language,
                                timeout: codeOptions.timeout,
                                memory: this.memory,
                                CPUs: this.CPUs
                            })];
                    case 1:
                        result = _a.sent();
                        return [2 /*return*/, result];
                }
            });
        });
    };
    Worker.prototype.build = function (langs) {
        return __awaiter(this, void 0, void 0, function () {
            var e_1;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        _a.trys.push([0, 2, , 3]);
                        return [4 /*yield*/, this.builder.build(langs)];
                    case 1:
                        _a.sent();
                        return [2 /*return*/, null];
                    case 2:
                        e_1 = _a.sent();
                        return [2 /*return*/, Promise.reject(e_1)];
                    case 3: return [2 /*return*/];
                }
            });
        });
    };
    Worker.prototype.start = function () {
        var _this = this;
        this.queue.process(function (job, done) { return __awaiter(_this, void 0, void 0, function () {
            var result, e_2;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        logger_1["default"].info("Received: " + job.data.id);
                        _a.label = 1;
                    case 1:
                        _a.trys.push([1, 3, , 4]);
                        return [4 /*yield*/, this.work(job.data)];
                    case 2:
                        result = _a.sent();
                        done(null, result);
                        logger_1["default"].debug(JSON.stringify(result));
                        return [3 /*break*/, 4];
                    case 3:
                        e_2 = _a.sent();
                        done(e_2);
                        logger_1["default"].error(e_2);
                        return [3 /*break*/, 4];
                    case 4: return [2 /*return*/];
                }
            });
        }); });
    };
    Worker.prototype.pause = function () {
        this.queue.pause();
    };
    Worker.prototype.resume = function () {
        this.queue.resume();
    };
    Worker.prototype.stop = function () {
        this.queue.close();
    };
    return Worker;
}());
exports["default"] = Worker;
